__global__ static void hst_cuda_nearest_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + threadIdx.x;
    const int idy = bidy + threadIdx.y;

    float4 all1, all2;

    float res = 0;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = bx + tidx;
    const float y = by + tidy;

    int minh1, minh2;

    __shared__  float   cache_subh[BLOCK_SIZE_X][2 * BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[2 * BLOCK_SIZE_Y];
    __shared__  float   cache[2 * BLOCK_SIZE_Y][24];

    const int prj = tidx >> 3;
    const int pos = BLOCK_SIZE_X + (tidx & 7);
//	int prj = tidx&1;
//	int pos = BLOCK_SIZE_X + (tidx >> 1);

    const int tidy2 = tidy << 1;
    const int tidy21 = tidy2 + 1;
    const int tidy2p = tidy2 + prj;

//#pragma unroll 1
    for (int proje=0; proje<num_proj; proje += 2 * BLOCK_SIZE_Y) {
	int proj = proje + tidy2;
	
        all1 = c_all[proj];
        all2 = c_all[proj + 1];
	minh1 = (int)(all1.z + bx * all1.x - by * all1.y + all1.w);
	minh2 = (int)(all2.z + bx * all2.x - by * all2.y + all2.w);
		
	cache_subh[tidx][tidy2] = all1.z + x * all1.x - minh1;
	cache_subh[tidx][tidy21] = all2.z + x * all2.x - minh2;
	
	if (tidx == 0) {
	    cache_sin[tidy2] = all1.y;
	    cache_sin[tidy21] = all2.y;
	}


	cache[tidy2][tidx] = tex2D(tex_projes, minh1 + tidx, proj);
	cache[tidy21][tidx] = tex2D(tex_projes, minh2 + tidx, proj + 1);
        cache[tidy2p][pos] = tex2D(tex_projes, (prj?minh2:minh1) + pos, proj + prj);

	__syncthreads();

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    int j = 2 * i;
	    minh1 = cache_subh[tidx][j] - y * cache_sin[j];
	    res += cache[j][minh1];

	    ++j;
	    minh1 = cache_subh[tidx][j] - y * cache_sin[j];
	    res += cache[j][minh1];
	}

	__syncthreads();

    }

    d_SLICE[ BLOCK_SIZE_X*gridDim.x*idy + idx ] = res;
}

__global__ static void hst_cuda_linear_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + threadIdx.x;
    const int idy = bidy + threadIdx.y;

    float4 all1, all2;

    float res = 0;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = bx + tidx;
    const float y = by + tidy;

    float minh;
    int minh1, minh2;

    __shared__  float   cache_subh[BLOCK_SIZE_X][2 * BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[2 * BLOCK_SIZE_Y];
    __shared__  float2   cache[2 * BLOCK_SIZE_Y][24];

    const int prj = tidx >> 3;
    const int pos = BLOCK_SIZE_X + (tidx & 7);
//	int prj = tidx&1;
//	int pos = BLOCK_SIZE_X + (tidx >> 1);

    const int tidy2 = tidy << 1;
    const int tidy21 = tidy2 + 1;
    const int tidy2p = tidy2 + prj;

//#pragma unroll 1
    for (int proje=0; proje<num_proj; proje += 2 * BLOCK_SIZE_Y) {
	float projf = proje + tidy2 + 0.5;
	
        all1 = c_all[proje + tidy2];
        all2 = c_all[proje + tidy2 + 1];
	minh1 = (int)(all1.z + bx * all1.x - by * all1.y + all1.w);
	minh2 = (int)(all2.z + bx * all2.x - by * all2.y + all2.w);
		
	cache_subh[tidx][tidy2] = all1.z + x * all1.x - minh1;
	cache_subh[tidx][tidy21] = all2.z + x * all2.x - minh2;
	
	if (tidx == 0) {
	    cache_sin[tidy2] = all1.y;
	    cache_sin[tidy21] = all2.y;
	}

	cache[tidy2][tidx].x = tex2D(tex_projes, minh1 + tidx, projf);
	cache[tidy21][tidx].x = tex2D(tex_projes, minh2 + tidx, projf + 1);
        cache[tidy2p][pos].x = tex2D(tex_projes, (prj?minh2:minh1) + pos, projf + prj);

	__syncthreads();

    	cache[tidy2][tidx].y = cache[tidy2][tidx + 1].x - cache[tidy2][tidx].x;
	cache[tidy21][tidx].y = cache[tidy21][tidx + 1].x - cache[tidy21][tidx].x;
	if (pos < 23) {
	    cache[tidy2p][pos].y = cache[tidy2p][pos + 1].x - cache[tidy2p][pos].x;
	} else {
	    cache[tidy2p][pos].y = 0;
	}

	
	__syncthreads();
	
#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    int j = 2 * i;
	    float2 v;
	    
	    minh = cache_subh[tidx][j] - y * cache_sin[j];
	    minh1 = (int)minh;
	    v = cache[j][minh1];
	    res += v.x + (minh - minh1) * v.y;

	    ++j;
	    minh = cache_subh[tidx][j] - y * cache_sin[j];
	    minh1 = (int)minh;
	    v = cache[j][minh1];
	    res += v.x + (minh - minh1) * v.y;
	}

	__syncthreads();

    }

    d_SLICE[ BLOCK_SIZE_X*gridDim.x*idy + idx ] = res;
}



__global__ static void hst_cuda_oversample4_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float res = 0;
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + threadIdx.x;
    const int idy = bidy + threadIdx.y;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = bx + tidx;
    const float y = by + tidy;


    __shared__  float   cache_subh[BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[BLOCK_SIZE_Y];
    __shared__  float   cache[BLOCK_SIZE_Y][6 * BLOCK_SIZE_X];
    
//#pragma unroll 8
    float projf = tidy + 0.5;
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
        float4 all = c_all[proje + tidy];
	int minh = (int)(all.z + bx * all.x - by * all.y + all.w);

	cache_subh[tidx][tidy] = 4 * (all.z + x * all.x - minh);
	if (tidx == 0) cache_sin[tidy] = 4 * all.y;


#pragma unroll 6
	for (int i = 0; i < 6; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos] = tex2D(tex_projes, minh + ((float)pos)/4, proje + projf);
	}

	__syncthreads();

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    minh = cache_subh[tidx][i] - y * cache_sin[i];
	    res += cache[i][minh];
	}


	__syncthreads();

    }

    d_SLICE[ BLOCK_SIZE_X*gridDim.x*idy + idx ] = res;
}
