/*
 This is attempt to combine vliw and dma approaches
*/

#define to_int4(h) {(int)(h.x), (int)(h.y), (int)(h.z), (int)(h.w)}
#define subh_cache(subh) cache[(int)(subh)]
__kernel void hst_backproject_direct_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const float4 mult = {0.f, 1.f, 2.f, 3.f};

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;

    const float bx = bidx + off_x;
    const float by = bidy + off_y;


    const int rtidy = (tidy - 8) * 2 + tidx / 8;
    
        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = rtidy % 4;
    const int sbidy = rtidy / 4;
    
    const int stidx = tidx % 8;
    const int stidy = 0; // tidx / 8;
    
    const int sidx = - (sbidx * 8 + stidx);
    const int sidy = sbidy * 8; // + 4 * stidy;

    const int idx = bidx - sidx;
    const int idy = bidy + sidy;
    
    const int slice_width = 2 * get_global_size(0);

    float4 res = {0.f, 0.f, 0.f, 0.f};
    float4 res2 = {0.f, 0.f, 0.f, 0.f};

    int buf = 0;

        if (tidy < 8) {
            float4 all = c_all[tidy];
	    float minh = all.z + bx * all.x - by * all.y + all.w;

#pragma unroll 12
	    for (int i = 0; i < 12; i++) {
	        int pos = i * BLOCK_SIZE + tidx;
	        cache[tidy * (12 * BLOCK_SIZE + 1) + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, tidy + 0.5f)).x;
	    }
	}

    int proje = 0;
    while (1) {
        if (tidy > 7) {
            const int line = (buf?8:0);
#pragma unroll 2
	    for (int i = 0; i < 8; i+=4) {
	        int prj = i + proje;   

	        float pos = 193.f * (line + i);
                float4 i4 = pos + 193.f * mult;

                float4 all1 = c_all[prj];
                float4 all2 = c_all[prj + 1];
                float4 all3 = c_all[prj + 2];
                float4 all4 = c_all[prj + 3];

                float4 allx = {-4 * all1.x, -4 * all2.x, -4 * all3.x, -4 * all4.x};
                float4 ally = {-4 * all1.y, -4 * all2.y, -4 * all3.y, -4 * all4.y};
                float4 allw = {-4 * all1.w, -4 * all2.w, -4 * all3.w, -4 * all4.w};



                //float subh = (-sidx * all.x - sidy * all.y) - all.w;
                //float subh = mad(193.f, line + i, mad(sidx, all.x, mad(sidy, all.y, all.w)));
                float4 h = i4 + mad(sidx, allx, mad(sidy, ally, allw));
                
                {
                float4 h1 = h;
                float4 h2 = mad(1.f, ally, h);
                float4 h3 = mad(2.f, ally, h);
                float4 h4 = mad(3.f, ally, h);

                int4 ih1 = to_int4(h1);
                int4 ih2 = to_int4(h2);
                int4 ih3 = to_int4(h3);
                int4 ih4 = to_int4(h4);

	        res.x += subh_cache(ih1.x);
	        res.y += subh_cache(ih2.x);
	        res.z += subh_cache(ih3.x);
	        res.w += subh_cache(ih4.x);

	        res.x += subh_cache(ih1.y);
	        res.y += subh_cache(ih2.y);
	        res.z += subh_cache(ih3.y);
	        res.w += subh_cache(ih4.y);

	        res.x += subh_cache(ih1.z);
	        res.y += subh_cache(ih2.z);
	        res.z += subh_cache(ih3.z);
	        res.w += subh_cache(ih4.z);

	        res.x += subh_cache(ih1.w);
	        res.y += subh_cache(ih2.w);
	        res.z += subh_cache(ih3.w);
	        res.w += subh_cache(ih4.w);
	        }
	        {
                float4 h1 = mad(4.f, ally, h);
                float4 h2 = mad(5.f, ally, h);
                float4 h3 = mad(6.f, ally, h);
                float4 h4 = mad(7.f, ally, h);

                int4 ih1 = to_int4(h1);
                int4 ih2 = to_int4(h2);
                int4 ih3 = to_int4(h3);
                int4 ih4 = to_int4(h4);


	        res2.x += subh_cache(ih1.x);
	        res2.y += subh_cache(ih2.x);
	        res2.z += subh_cache(ih3.x);
	        res2.w += subh_cache(ih4.x);

	        res2.x += subh_cache(ih1.y);
	        res2.y += subh_cache(ih2.y);
	        res2.z += subh_cache(ih3.y);
	        res2.w += subh_cache(ih4.y);

	        res2.x += subh_cache(ih1.z);
	        res2.y += subh_cache(ih2.z);
	        res2.z += subh_cache(ih3.z);
	        res2.w += subh_cache(ih4.z);

	        res2.x += subh_cache(ih1.w);
	        res2.y += subh_cache(ih2.w);
	        res2.z += subh_cache(ih3.w);
	        res2.w += subh_cache(ih4.w);
                }
/*
            
	        res.x += subh_cache(subh);
	        res.y += subh_cache(mad(1.f, all.y, subh));
	        res.z += subh_cache(mad(2.f, all.y, subh));
	        res.w += subh_cache(mad(3.f, all.y, subh));

	        res2.x += subh_cache(mad(4.f, all.y, subh));
	        res2.y += subh_cache(mad(5.f, all.y, subh));
	        res2.z += subh_cache(mad(6.f, all.y, subh));
	        res2.w += subh_cache(mad(7.f, all.y, subh));
*/
	    }
        }
        
        proje += 8;
        if (proje >= num_proj) break;
        buf = !buf;

        if (tidy < 8) {
	    const int line = tidy + (buf?8:0);
	    int proj = proje + tidy;
            float4 all = c_all[proj];
	    float minh = all.z + bx * all.x - by * all.y + all.w;

#pragma unroll 12
	    for (int i = 0; i < 12; i++) {
	        int pos = i * BLOCK_SIZE + tidx;
	        cache[line * (12 * BLOCK_SIZE + 1) + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, proj + 0.5f)).x;
	    }
	}
        
        barrier(CLK_LOCAL_MEM_FENCE);
    }

    if (tidy > 7) {
        slice[slice_width * (idy    ) + idx] = res.x;
        slice[slice_width * (idy + 1) + idx] = res.y;
        slice[slice_width * (idy + 2) + idx] = res.z;
        slice[slice_width * (idy + 3) + idx] = res.w;

        slice[slice_width * (idy + 4) + idx] = res2.x;
        slice[slice_width * (idy + 5) + idx] = res2.y;
        slice[slice_width * (idy + 6) + idx] = res2.z;
        slice[slice_width * (idy + 7) + idx] = res2.w;
    }

}
