/*
 Attempt to use texture engine and ALUs simulatenously. Here, we try to exploit
 the VLIW capabilities of older AMD cards. There is sequential computations 
 and texture fetches on independent data.
*/

#define BLOCK_STEP 16

#define subh_cache(i, subh) cache[i * 12 * BLOCK_SIZE + (int)(subh)]
__kernel void hst_backproject_direct_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int tidx2 = 2 * get_local_id(0);
    const int tidy2 = 2 * get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;

        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
#if BLOCK_STEP == 1
    const int sidx = sbidx * 4 + stidx;
    const int sidy = sbidy * 4 + stidy;
    const int sidx2 = 2 * sidx;
    const int sidy2 = 2 * sidy;
#else    
    const int sidx = tidx;
    const int sidy = tidy;
    const int sidx2 = sidx;
    const int sidy2 = sidy;
#endif

    const int idx = bidx + sidx2;//2 * get_global_id(0);
    const int idy = bidy + sidy2;//2 * get_global_id(1);

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

//    const float x = idx + off_x;
//    const float y = idy + off_y;

    const float x = idx + off_x;
    const float y = -(idy + off_y);

    const int slice_width = 2 * get_global_size(0);
    
//    const int mode = ((sbidx < 3)&&(sbidy < 3))?1:0;
//    const int mode = (tidy > 7)?1:0;

    float2 res[2] = {{0.f, 0.f}, {0.f, 0.f}};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	float minh = all.z + bx * all.x - by * all.y + all.w;

#pragma unroll 12
	for (int i = 0; i < 12; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * 12 * BLOCK_SIZE + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
        for (int i = 0; i < BLOCK_SIZE; i++) {
//                float4 all = 4 * c_all[proje + i];

	        int proj = proje + i;
	        float projf = proj + 0.5;
                float4 all = c_all[proj];

                float subh = (sidx2 * all.x - sidy2 * all.y) - all.w;
                float subhx = subh + BLOCK_STEP * all.x;
//                float hx = all.z + x * all.x + y * all.y + all.x;
                float h = mad(y, all.y, mad(x, all.x, all.z));
                float hx = h + BLOCK_STEP * all.x;

	        res[0].x += subh_cache(i, 4 * subh);
//	        res[0].x += read_imagef(sinogram, volumeSampler, (float2)(h, projf)).x;
	        res[0].y += subh_cache(i, 4 * subhx);
//	        res[0].y += read_imagef(sinogram, volumeSampler, (float2)(hx, projf)).x;
//	        res[1].x += subh_cache(i, 4 * (subh - BLOCK_STEP * all.y));
	        res[1].x += read_imagef(sinogram, volumeSampler, (float2)(h - BLOCK_STEP * all.y, projf)).x;
//	        res[1].y += subh_cache(i, 4 * (subhx - BLOCK_STEP * all.y));
	        res[1].y += read_imagef(sinogram, volumeSampler, (float2)(hx - BLOCK_STEP *  all.y, projf)).x;
        }
        
        barrier(CLK_LOCAL_MEM_FENCE);
    }
    
    slice[slice_width * (idy    ) + idx] = res[0].x;
    slice[slice_width * (idy    ) + idx + BLOCK_STEP] = res[0].y;
    slice[slice_width * (idy + BLOCK_STEP) + idx] = res[1].x;
    slice[slice_width * (idy + BLOCK_STEP) + idx + BLOCK_STEP] = res[1].y;
}
